unit DLASim;

interface

uses
  Windows, SysUtils, Classes, Types, Math, Graphics, RejbrandCommon;

type
  PBitmap = ^TBitmap;
  TBitmap = class(TObject)
  public
    type
      TPixel = type TColor;
    const
      BLACK = TPixel($00000000);
      WHITE = TPixel($FFFFFFFF); // SIC! It is important that the
                                 // reserved byte is also $FF.
  private
    FBitmap: array of array of TPixel;
    function GetHeight: integer;
    function GetWidth: integer;
    procedure SetHeight(const Value: integer);
    procedure SetWidth(const Value: integer);
    function GetPixel(X, Y: integer): TPixel;
    procedure SetPixel(X, Y: integer; const Value: TPixel);
  public
    procedure FillWhite;
    procedure FillBlack;
    procedure SetSize(const AWidth, AHeight: integer);
    function PixelExists(const APixel: TPoint): boolean; overload;
    function PixelExists(const AX, AY: integer): boolean; overload;
    function CreateASPixmap: TASPixmap;
    function CreateGDIBitmap: Graphics.TBitmap;
    property Pixels[X, Y: integer]: TPixel read GetPixel write SetPixel;
    property Width: integer read GetWidth write SetWidth;
    property Height: integer read GetHeight write SetHeight;
  end;

type
  PDLACreateSettings = ^TDLACreateSettings;
  TDLACreateSettings = record
    Size: integer;
    ThreadCount: integer;
    MaxNumMoleculesAdsorbed,
    MaxNumMoleculesUsed: integer;
    UsedMolecules,
    AdsorbedMolecules: integer;
    Radius,
    DurationInSecs: real;
    RadiusCriticalSection: TRTLCriticalSection;
  end;

  TDLAMode = (dlaCircularTrue, dlaCircularFast, dlaUniform, dlaBoundary);

  PDLAFractal = ^TDLAFractal;
  TDLAFractal = class(TBitmap)
  private
    FMode: TDLAMode;
    FCircularCompensator: boolean;
    FThreadCount: integer;
    FCachedCX,
    FCachedCY: integer;
    FThreads: array of THandle;
    FCachedRadius: real;
    FCachedParticleCount: integer;
    FCachedRadiusOfInscribedCircle: integer;
    FRadiusCriticalSection: TRTLCriticalSection;
    FTerminatorCriticalSection: TRTLCriticalSection;
    FTerminate: boolean;
    FRunning: boolean;
    function GetCX: integer;
    function GetCY: integer;
    procedure FixWhite;
    function GetRadiusOfInscribedCircle: integer;
  public
    procedure LoadFromFile(const FileName: string);
    procedure SaveToFile(const FileName: string);
    procedure DrawPointAtCentre;
    procedure DrawRandomPoints(const ANumber: integer);
    procedure DrawCircle(const ARadius: integer);
    procedure DrawBoundary;
    procedure DrawLineAtMiddle;
    procedure DrawLineAtBottom;
    function GetRadius: real;
    function GetParticleCount: integer;
    procedure Simulate;
    procedure Pause;
    constructor Create;
    destructor Destroy; override;
    property Mode: TDLAMode read FMode write FMode default dlaCircularTrue;
    property CircularCompensator: boolean read FCircularCompensator write FCircularCompensator default false;
    property ThreadCount: integer read FThreadCount write FThreadCount default 1;
    property Running: boolean read FRunning;
    property Radius: real read GetRadius;
    property ParticleCount: integer read GetParticleCount;
    property Width;
    property Height;
  end;

function CreateDLAFractal(var Settings: TDLACreateSettings): TBitmap;

implementation

{ TBitmap }

function TBitmap.CreateASPixmap: TASPixmap;
var
  y: Integer;
begin
  Assert(sizeof(TPixel) = 4);
  SetLength(result, Height, Width);
  for y := 0 to Height - 1 do
    Move(FBitmap[y, 0], result[y, 0], Width * sizeof(TPixel));
end;

function TBitmap.CreateGDIBitmap: Graphics.TBitmap;
var
  y: Integer;
begin
  Assert(sizeof(TPixel) = 4);
  result := Graphics.TBitmap.Create;
  result.SetSize(Width, Height);
  result.PixelFormat := pf32bit;
  for y := 0 to Height - 1 do
    Move(FBitmap[y, 0], result.ScanLine[y]^, Width * sizeof(TPixel));
end;

procedure TBitmap.FillWhite;
var
  y: Integer;
begin
  for y := 0 to Height - 1 do
    FillChar(FBitmap[y, 0], Width * sizeof(TPixel), $FF);
end;

procedure TBitmap.FillBlack;
var
  y: Integer;
begin
  for y := 0 to Height - 1 do
    FillChar(FBitmap[y, 0], Width * sizeof(TPixel), $00);
end;

function TBitmap.GetHeight: integer;
begin
  result := length(FBitmap);
end;

function TBitmap.GetPixel(X, Y: integer): TPixel;
begin
  result := FBitmap[Y, X];
end;

function TBitmap.GetWidth: integer;
begin
  if FBitmap <> nil then
    result := length(FBitmap[0])
  else
    result := 0;
end;

function TBitmap.PixelExists(const APixel: TPoint): boolean;
begin
  result := InRange(APixel.X, 0, GetWidth - 1) and
    InRange(APixel.Y, 0, GetHeight - 1);
end;

function TBitmap.PixelExists(const AX, AY: integer): boolean;
begin
  result := InRange(AX, 0, GetWidth - 1) and InRange(AY, 0, GetHeight - 1);
end;

procedure TBitmap.SetHeight(const Value: integer);
begin
  SetLength(FBitmap, Value, GetWidth);
end;

procedure TBitmap.SetPixel(X, Y: integer; const Value: TPixel);
begin
  FBitmap[Y, X] := Value;
end;

procedure TBitmap.SetSize(const AWidth, AHeight: integer);
begin
  SetLength(FBitmap, AHeight, AWidth);
end;

procedure TBitmap.SetWidth(const Value: integer);
begin
  SetLength(FBitmap, GetHeight, Value);
end;

function RandomAngle: real; inline;
begin
  result := 2*Pi*Random;
end;

procedure DoRandomStep(var APoint: TPoint); inline;
begin
  case RandomRange(0, 4) of
    0: inc(APoint.X);
    1: dec(APoint.X);
    2: inc(APoint.Y);
    3: dec(APoint.Y);
  end;
end;

type
  PDLAData = ^TDLAData;
  TDLAData = record
    Settings: PDLACreateSettings;
    Bitmap: PBitmap;
    cx, cy: integer;
  end;

function RandomWalker(Parameter: Pointer): integer;
var
  pnt: TPoint;
  sintheta, costheta: extended;
  cx, cy: integer;

  Settings: PDLACreateSettings;
  Bitmap: PBitmap;

  rCache: real;

  function WithinCircle: boolean;
  begin
    result := Hypot(pnt.X - cx, pnt.Y - cy) < rCache + 6;
  end;

  function ShouldAdsorb: boolean;
  var
    i: Integer;
    j: Integer;
  begin
    result := false;
    if not Bitmap.PixelExists(pnt.X, pnt.Y) then Exit;
    for i := -1 to 1 do
      for j := -1 to 1 do
        if Bitmap.PixelExists(pnt.X + i, pnt.Y + j) and
          (Bitmap.Pixels[pnt.X + i, pnt.Y + j] <> TBitmap.WHITE) then
          Exit(true);
  end;

begin
  Settings := PDLAData(Parameter).Settings;
  Bitmap := PDLAData(Parameter).Bitmap;

  cx := PDLAData(Parameter).cx;
  cy := PDLAData(Parameter).cy;

  while (Settings.AdsorbedMolecules < Settings.MaxNumMoleculesAdsorbed) and
    (Settings.UsedMolecules < Settings.MaxNumMoleculesUsed) do
  begin
    InterlockedIncrement(Settings.UsedMolecules);
    SinCos(RandomAngle, sintheta, costheta);
    EnterCriticalSection(Settings.RadiusCriticalSection);
      rCache := Settings.Radius;
    LeaveCriticalSection(Settings.RadiusCriticalSection);
    pnt.X := Round(cx + rCache*costheta);
    pnt.Y := Round(cy + rCache*sintheta);
    while WithinCircle do
    begin
      if ShouldAdsorb then
      begin
        Bitmap.Pixels[pnt.X, pnt.Y] := TBitmap.BLACK;
        InterlockedIncrement(Settings.AdsorbedMolecules);
        EnterCriticalSection(Settings.RadiusCriticalSection);
          Settings.Radius := Max(Settings.Radius, hypot(pnt.X - cx, pnt.Y - cy));
        LeaveCriticalSection(Settings.RadiusCriticalSection);
        break;
      end;
      DoRandomStep(pnt);
    end;
  end;

  result := 0;

end;

function CreateDLAFractal(var Settings: TDLACreateSettings): TBitmap;
var
  data: TDLAData;
  c1, c2, f: Int64;
  dummy: cardinal;
  i: Integer;
  threads: array of THandle;
begin
  QueryPerformanceCounter(c1);
  QueryPerformanceFrequency(f);

  result := TBitmap.Create;
  result.SetSize(Settings.Size, Settings.Size);
  result.FillWhite;

  Settings.UsedMolecules := 0;
  Settings.AdsorbedMolecules := 0;
  Settings.Radius := 6;

  data.Settings := @Settings;
  data.Bitmap := @result;
  data.cx := Settings.Size div 2;
  data.cy := Settings.Size div 2;

  result.Pixels[data.cx, data.cy] := TBitmap.BLACK;

  InitializeCriticalSection(Settings.RadiusCriticalSection);

  SetLength(threads, Settings.ThreadCount);
  for i := 0 to Settings.ThreadCount - 1 do
    threads[i] := BeginThread(nil, 0, @RandomWalker, @Data, 0, dummy);

  if WaitForMultipleObjects(Settings.ThreadCount, @threads[0], true, INFINITE) = WAIT_FAILED then
    RaiseLastOSError;

  DeleteCriticalSection(Settings.RadiusCriticalSection);

  QueryPerformanceCounter(c2);
  Settings.DurationInSecs := (c2-c1) / f;

end;

{ TDLAFractal }

constructor TDLAFractal.Create;
begin
  inherited;
  FMode := dlaCircularTrue;
  FCircularCompensator := false;
  FThreadCount := 1;
  FCachedRadius := 6;
  FCachedParticleCount := 0;
  InitializeCriticalSection(FTerminatorCriticalSection);
  InitializeCriticalSection(FRadiusCriticalSection);
  FTerminate := false;
  FRunning := false;
end;

destructor TDLAFractal.Destroy;
begin
  DeleteCriticalSection(FRadiusCriticalSection);
  DeleteCriticalSection(FTerminatorCriticalSection);
  inherited;
end;

procedure TDLAFractal.DrawBoundary;
var
  i: Integer;
begin
  for i := 0 to Width - 1 do
  begin
    Pixels[i, 0] := TBitmap.BLACK;
    Pixels[i, Height - 1] := TBitmap.BLACK;
  end;
  for i := 0 to Height - 1 do
  begin
    Pixels[0, i] := TBitmap.BLACK;
    Pixels[Width - 1, i] := TBitmap.BLACK;
  end;
end;

procedure TDLAFractal.DrawCircle(const ARadius: integer);
var
  cx, cy: integer;
  x, y: integer;
  cosphi, sinphi: extended;
  t: Integer;
begin
  cx := GetCX;
  cy := GetCY;
  for t := 0 to ceil(2*Pi*ARadius - 1) do
  begin
    SinCos(t / ARadius, sinphi, cosphi);
    x := Round(cx + ARadius*cosphi);
    y := Round(cy + ARadius*sinphi);
    if PixelExists(x, y) then
      Pixels[x, y] := TBitmap.BLACK;
  end;

end;

procedure TDLAFractal.DrawLineAtBottom;
var
  i: Integer;
begin
  for i := 0 to Width - 1 do
    Pixels[i, Height - 1] := TBitmap.BLACK;
end;

procedure TDLAFractal.DrawLineAtMiddle;
var
  i: Integer;
  cy: integer;
begin
  cy := GetCY;
  for i := 0 to Width - 1 do
    Pixels[i, cy] := TBitmap.BLACK;
end;

procedure TDLAFractal.DrawPointAtCentre;
begin
  if not Assigned(FBitmap) then Exit;
  Pixels[GetCX, GetCY] := TBitmap.BLACK;
end;

procedure TDLAFractal.DrawRandomPoints(const ANumber: integer);
var
  i: Integer;
begin
  if not Assigned(FBitmap) then Exit;
  for i := 0 to ANumber - 1 do
    Pixels[Random(Width), Random(Height)] := TBitmap.BLACK;
end;

function TDLAFractal.GetCX: integer;
begin
  result := Width div 2;
  FCachedCX := result;
end;

function TDLAFractal.GetCY: integer;
begin
  result := Height div 2;
  FCachedCY := result;
end;

function TDLAFractal.GetParticleCount: integer;
var
  y: Integer;
  x: Integer;
begin
  result := 0;
  for y := 0 to Height - 1 do
    for x := 0 to Width - 1 do
      if Pixels[x, y] <> TBitmap.WHITE then
        inc(result);
  FCachedParticleCount := result;
end;

function TDLAFractal.GetRadius: real;
var
  cx, cy: integer;
  y: Integer;
  x: Integer;
begin
  cx := GetCX;
  cy := GetCY;
  result := 6; // SIC!
  for y := 0 to Height - 1 do
    for x := 0 to Width - 1 do
      if Pixels[x, y] <> TBitmap.WHITE then
        result := Max(result, hypot(x - cx, y - cy));
  FCachedRadius := result;
end;

function TDLAFractal.GetRadiusOfInscribedCircle: integer;
begin
  result := Min(Width, Height) div 2;
  FCachedRadiusOfInscribedCircle := result;
end;

procedure TDLAFractal.FixWhite;
var
  y: Integer;
  x: Integer;
begin
  // Ugly but necessary.
  for y := 0 to Height - 1 do
    for x := 0 to Width - 1 do
      if (Pixels[x, y] and $00FFFFFF) = $FFFFFF then
        Pixels[x, y] := TBitmap.WHITE;

end;

procedure TDLAFractal.LoadFromFile(const FileName: string);
var
  y: integer;
begin
  Assert(sizeof(TPixel) = 4);
  with Graphics.TBitmap.Create do
    try
      LoadFromFile(FileName);
      Self.SetSize(Width, Height);
      PixelFormat := pf32bit;
      for y := 0 to Height - 1 do
        Move(Scanline[y]^, FBitmap[y, 0], Width * sizeof(TPixel));
    finally
      Free;
    end;
  FixWhite;
end;

procedure TDLAFractal.Pause;
begin
  EnterCriticalSection(FTerminatorCriticalSection);
  FTerminate := true;
  LeaveCriticalSection(FTerminatorCriticalSection);
  if WaitForMultipleObjects(FThreadCount, @FThreads[0], true, INFINITE) = WAIT_FAILED then
    RaiseLastOSError;
  FTerminate := false;
  FRunning := false;
end;

function DLARandomWalker(Parameter: Pointer): integer;
var
  pnt: TPoint;
  sintheta, costheta: extended;
  cx, cy: integer;

  DLAFractal: TDLAFractal;

  rCache: real;

  function WithinRegion: boolean;
  begin
    case DLAFractal.Mode of
      dlaCircularTrue:
        result := Hypot(pnt.X - cx, pnt.Y - cy) < DLAFractal.FCachedRadiusOfInscribedCircle + 6;
      dlaCircularFast:
        result := Hypot(pnt.X - cx, pnt.Y - cy) < rCache + 6;
      dlaUniform:
        result := DLAFractal.PixelExists(pnt);
      dlaBoundary:
        result := DLAFractal.PixelExists(pnt);
    end;
  end;

  function WithinInscribedCircle: boolean;
  begin
    result := Hypot(pnt.X - cx, pnt.Y - cy) < DLAFractal.FCachedRadiusOfInscribedCircle;
  end;

  function ShouldAdsorb: boolean;
  var
    i: Integer;
    j: Integer;
  begin
    result := false;
    if not DLAFractal.PixelExists(pnt.X, pnt.Y) then Exit;
    if DLAFractal.CircularCompensator and not WithinInscribedCircle then Exit;
    for i := -1 to 1 do
      for j := -1 to 1 do
        if DLAFractal.PixelExists(pnt.X + i, pnt.Y + j) and
          (DLAFractal.Pixels[pnt.X + i, pnt.Y + j] <> TBitmap.WHITE) then
          Exit(true);
  end;

  function IsTerminated: boolean;
  begin
    EnterCriticalSection(DLAFractal.FTerminatorCriticalSection);
    result := DLAFractal.FTerminate;
    LeaveCriticalSection(DLAFractal.FTerminatorCriticalSection);
  end;

begin
  DLAFractal := TDLAFractal(Parameter);

  cx := DLAFractal.FCachedCX;
  cy := DLAFractal.FCachedCY;

  while not IsTerminated do
  begin

    // Set initial position
    case DLAFractal.Mode of
      dlaCircularTrue:
        begin
          SinCos(RandomAngle, sintheta, costheta);
          pnt.X := Round(cx + DLAFractal.FCachedRadiusOfInscribedCircle*costheta);
          pnt.Y := Round(cy + DLAFractal.FCachedRadiusOfInscribedCircle*sintheta);
        end;
      dlaCircularFast:
        begin
          SinCos(RandomAngle, sintheta, costheta);
          EnterCriticalSection(DLAFractal.FRadiusCriticalSection);
            rCache := DLAFractal.FCachedRadius;
          LeaveCriticalSection(DLAFractal.FRadiusCriticalSection);
          pnt.X := Round(cx + rCache*costheta);
          pnt.Y := Round(cy + rCache*sintheta);
        end;
      dlaUniform:
        begin
          pnt.X := Random(DLAFractal.Width);
          pnt.Y := Random(DLAFractal.Height);
        end;
      dlaBoundary:
        begin
          case Random(4) of
            0: pnt := Point(Random(DLAFractal.Width), 6);
            1: pnt := Point(DLAFractal.Width - 1 - 6, Random(DLAFractal.Height));
            2: pnt := Point(Random(DLAFractal.Width), DLAFractal.Height - 6);
            3: pnt := Point(6, Random(DLAFractal.Height));
          end;
        end;
    end;

    // Walk
    while WithinRegion do
    begin
      if ShouldAdsorb then
      begin
        DLAFractal.Pixels[pnt.X, pnt.Y] := TBitmap.BLACK;
        InterlockedIncrement(DLAFractal.FCachedParticleCount);
        EnterCriticalSection(DLAFractal.FRadiusCriticalSection);
          DLAFractal.FCachedRadius := Max(DLAFractal.FCachedRadius,
            hypot(pnt.X - cx, pnt.Y - cy));
        LeaveCriticalSection(DLAFractal.FRadiusCriticalSection);
        break;
      end;
      DoRandomStep(pnt);
    end;

  end;

  result := 0;
end;

procedure TDLAFractal.SaveToFile(const FileName: string);
begin
  with CreateGDIBitmap do
    try
      SaveToFile(FileName);
    finally
      Free;
    end;
end;

procedure TDLAFractal.Simulate;
var
  i: integer;
  dummy: cardinal;
begin

  GetRadiusOfInscribedCircle;

  GetRadius; //
  GetCX;     // update the cached values
  GetCY;     //

  FTerminate := false;
  SetLength(FThreads, FThreadCount);
  for i := 0 to FThreadCount - 1 do
    FThreads[i] := BeginThread(nil, 0, @DLARandomWalker, Self, 0, dummy);

  FRunning := true;
end;

initialization
  Randomize;

end.
